/**
 * Copyright (c) 2014 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather;

import java.util.ArrayList;
import java.util.Date;
import java.util.Set;

import javax.swing.SwingUtilities;

import weather.bom.ReportParser;

import weather.db.DBUtils;
import weather.db.LockManager;
import weather.db.ObservationReader;
import weather.db.ObservationWriter;
import weather.db.ObservationCache;

import weather.gui.MainWindow;
import weather.gui.PlottingPanel;
import weather.gui.DetailsDialog;
import weather.gui.ActionsDialog;

import weather.selector.Selector;


/**
 * The controller for the weather application.
 * This handles the logic of the application.
 */
public class Controller
{
	private MainWindow mMainWindow;
	private PlottingPanel mPlottingPanel;
	private DetailsDialog mDetailsDialog;
	private ActionsDialog mActionsDialog;

	private String mSiteKey;
	private String mURL;
	private Selector mSelector1;
	private Selector mSelector2;
	private Report mReport;
	private String mSiteName;

	private ObservationList mFullObservationList;
	private ObservationList mViewedObservationList;
	private ObservationList mObservationList1;
	private ObservationList mObservationList2;

	private Observation mSelectedObservation;
	private boolean mSelectedIsPlot1;

	private String mMarkerType;
	private Observation mMarkerObservation;

	private final ObservationCache mObservationCache = new ObservationCache();
	private final Configuration mConfig = Configuration.getInstance();
	private TimeIntervals mTimeIntervals;

	/**
	 * The maximum time difference in milliseconds between the
	 * first and last observation to show.  A value of zero indicates
	 * that only the currently collected report is to be shown.
	 * A value of MAX_VALUE indicates that all observations for the
	 * site are to be shown.
	 */
	private long mTimeInterval;

	/**
	 * The shift value for the plot.
	 * This is the number of hours to shift the view leftwards by.
	 * The value should be a multiple of the major time interval.
	 * This is ignored if only the currently collected report is shown.
	 */
	private int mShiftHours;

	//
	// The major and minor time intervals for the plot in minutes.
	// These values are calculated from the size of the time interval.
	//
	private int mTimeMajorInterval;
	private int mTimeMinorInterval;

	//
	// The minimum and maximum times for the plot.
	// These aren't the same as the times of the first or last observations
	// since rounding to the major time interval is done and there may
	// be gaps in the observations.
	//
	private DateTime mMinimumTime;
	private DateTime mMaximumTime;

	/**
	 * The time in milliseconds to sleep between the automatic updates.
	 */
	private long mSleepTime;

	/**
	 * The number of hours in an interval when displaying data in the
	 * grouping mode.  In this mode the maximum and minimum values
	 * are shown for each interval.  If this is zero then the normal plot is
	 * done.  Otherwise the value must either be an exact divisor of a
	 * day or else one of several special multiple-day intervals.
	 */
	private int mGroupHours;

	/**
	 * Whether uniform scaling is being used.
	 */
	private boolean mIsUniformScaling;

	/**
	 * Whether UTC times are shown.
	 */
	private boolean mIsShowUTCTimes;

	/**
	 * The time that the current plot is being started.
	 * This value is used for determining the beginning and ending times
	 * for the plot, and is never before the latest observation's time.
	 */
	private DateTime mPlotTime;


	/**
	 * The constructor for the controller.
	 * We obtain any configuration values we need.
	 */
	public Controller()
	{
		//
		// Get the initial time interval from the configuration file.
		// converting from days to milliseconds.
		//
		mTimeInterval = mConfig.getTimeIntervalDays() * DateTime.DAY_INTERVAL;

		//
		// Get some other values from the configuration.
		//
		mIsUniformScaling = mConfig.isUniformScaling();
		mIsShowUTCTimes = mConfig.isShowUTCTimes();
		mMarkerType = mConfig.getMarkerType();

		String timeIntervalsString = mConfig.getTimeIntervals();

		mTimeIntervals = new TimeIntervals(timeIntervalsString);

		//
		// Initialise the selectors for the air temperature.
		//
		mSelector1 = Selector.AIR_TEMP_SELECTOR;
	}


	public void setMainWindow(MainWindow mainWindow)
	{
		mMainWindow = mainWindow;
	}


	public void setPlottingPanel(PlottingPanel plottingPanel)
	{
		mPlottingPanel = plottingPanel;
	}


	public void setDetailsDialog(DetailsDialog detailsDialog)
	{
		mDetailsDialog = detailsDialog;
	}


	public void setActionsDialog(ActionsDialog actionsDialog)
	{
		mActionsDialog = actionsDialog;
	}


	/**
	 * Set the report to be displayed according to the specified site
	 * key which identifies the site according to the configuration file.
	 * The site key does not have the "Site_" prefix.  Nothing is done
	 * if the site key is null, empty, or undefined.
	 * @param letter
	 */
	public void setReport(String siteKey)
	{
		if ((siteKey == null) || siteKey.isEmpty())
			return;

		String URL = mConfig.getSiteURL(siteKey);

		if (URL == null)
			return;

		mURL = URL;
		mSiteKey = siteKey;

		mActionsDialog.setSelectedSite(mSiteKey);

		mSelectedObservation = null;
		mSelectedIsPlot1 = true;

		updateReport();
	}


	/**
	 * Collect and display the latest report from the currently selected site.
	 * This displays a busy cursor while the report is being collected.
	 */
	public void updateReport()
	{
		mMainWindow.showBusy(true);

		try
		{
			mReport = collectReport(mURL);

			if (mReport == null)
				return;

			mSiteName = mReport.getSiteName();

			//
			// Update the database using the new report.
			//
			updateDatabase(mReport);

			//
			// Update the plot with the new data with
			// no shifting of the view to begin with.
			//
			mShiftHours = 0;

			updatePlot(true);
		}
		catch (Exception e)
		{
			Utils.logError("Exception reading report", e);

			return;
		}
		finally
		{
			mMainWindow.showBusy(false);
		}
	}


	/**
	 * Try to collect the latest weather report from the specified URL.
	 * Returns null on an error.
	 * @param URL
	 * @return
	 */
	private Report collectReport(String URL)
	{
		try
		{
			ReportCollector collector = new ReportCollector();

			String reportText = collector.collectHTTPReport(URL);

			ReportParser parser = new ReportParser();

			return parser.parse(reportText);
		}
		catch (Exception e)
		{
			Utils.logError("Exception reading report from " + URL, e);

			return null;
		}
	}


	/**
	 * Update the current plot using the report that was last collected.
	 * This is used after changing major plotting parameters such as the
	 * time interval or the view.  The list of observations from the current
	 * report can be truncated or expanded to include older observations
	 * from the database according to the current time interval for plotting.
	 * The isFullUpdate flag is true if the site is changed or the time
	 * intervals have changed, in which case cached data must be discarded.
	 */
	private void updatePlot(boolean isFullUpdate)
	{
		//
		// Get the time to be used for the plot based on the last
		// observation time or the current time, whichever is greater.
		//
		mPlotTime = calculatePlotTime(mSiteName);

		//
		// Get the full list of observations for the site.
		//
		mFullObservationList = mObservationCache.extractSiteObservations(mSiteName);

		//
		// Start with the report's original observations.
		//
		mViewedObservationList = mReport.getObservationList();

		//
		// If the current report's observations aren't wanted then
		// use the observations from the full list instead.
		//
		if (mTimeInterval > 0)
			mViewedObservationList = mFullObservationList;
			
		//
		// Calculate the time-related values needed for the plot.
		//
		calculateTimeValues();

		//
		// If the time interval is not special then extract the observations
		// from the full list for just the required interval.
		//
		if ((mTimeInterval > 0) && (mTimeInterval != Long.MAX_VALUE))
		{
			mViewedObservationList = mFullObservationList.extractObservations(
				mMinimumTime, mMaximumTime);
		}

		//
		// If we are not plotting intervals then set both of the observation lists
		// for the plot to be the same as the viewed list.  Otherwise prune the
		// observation lists based on the maximums and minimums for the
		// specified intervals.
		//
		mObservationList1 = mViewedObservationList;
		mObservationList2 = mViewedObservationList;

		if (mGroupHours > 0)
		{
			mSelector2 = mSelector1;

			mObservationList1 = mObservationList1.extractMaximumsFromIntervals(
				mSelector1, mGroupHours, mIsShowUTCTimes);

			mObservationList2 = mObservationList2.extractMinimumsFromIntervals(
				mSelector2, mGroupHours, mIsShowUTCTimes);
		}

		//
		// Find the marker observer if any for the plot.
		//
		findMarkerObservation();

		//
		// Create the set of parameters and give it to the plotting panel to display.
		//
		PlotParameters plotParameters  = createPlotParameters();

		mPlottingPanel.updatePlot(plotParameters, isFullUpdate);
	}


	/**
	 * Update the database cache and data files with the possibly new
	 * observations in the specified report which was collected from
	 * one of the weather sites.
	 * @param observationList
	 */
	private void updateDatabase(Report report)
	{
		mSiteName = report.getSiteName();

		ObservationList observationList = report.getObservationList();

		if (observationList.isEmpty())
			return;

		//
		// Lock the database for exclusive access while we access it.
		// We try to proceed anyway if it fails since exactness isn't critical.
		//
		LockManager.getLock();

		//
		// Read any database files into the observation cache for the site.
		//
		readOldObservations(mSiteName);

		//
		// Now add any new observations to the observation cache.
		// Observations already in the database (and so now also in
		// the observation cache) will not be returned.  The list of
		// new observations is returned.
		//
		ObservationList newList =
			mObservationCache.addObservationList(observationList);

		//
		// If there were any new observations collected then save them
		// into the database.
		//
		writeNewObservations(newList);

		//
		// Now we can unlock the database.
		//
		LockManager.freeLock();
	}


	/**
	 * Write the specified new observations (if any) to the database.
	 * This catches exceptions.
	 */
	private void writeNewObservations(ObservationList observationList)
	{
		if (observationList.isEmpty())
			return;

		try
		{
			ObservationWriter writer = new ObservationWriter();

			writer.writeObservations(observationList);
		}
		catch (Exception e)
		{
			Utils.logError("Exception writing observations to database", e);
		}
	}


	/**
	 * Read any old observations from the database (if necessary).
	 */
	private void readOldObservations(String siteName)
	{
		//
		// Get a list of the year/month number values for all the data
		// files of the specified site name.  The returned numbers are
		// of the form 100 * year + month.
		//
		ArrayList<Integer> yearMonthList =
			DBUtils.getDataFileNumbers(siteName);

		//
		// Process each of the returned year/month values.
		//
		for (int number : yearMonthList)
		{
			int year = number / 100;
			int month = number % 100;

			readOldObservations(siteName, year, month);
		}
	}


	/**
	 * Read any observations for the specified site, year, and month from
	 * the database (if necessary) and put them into the observation cache.
	 * The database file is read if the cache data is not present or if the
	 * cached data is not current.  Having putting them there, any duplicate
	 * observations read from a new report won't be treated as new and
	 * written again to the database.
	 * @param siteName
	 * @param year
	 * @param month
	 */
	private void readOldObservations(String siteName, int year, int month)
	{
		//
		// If the cache data exists and is still current then there is
		// nothing to do.
		//
		if (mObservationCache.isCurrent(siteName, year, month))
			return;

		//
		// Get the observation reader and try to read the observations
		// from the database for this combination.  An empty list is
		// returned if the read fails (usually file not found).
		//
		ObservationReader reader = new ObservationReader();

		ObservationList observationList =
			reader.readObservations(siteName, year, month);

		//
		// Get the last modified time of the file and store that into
		// the cache entry for this site and time to make it current.
		//
		long lastModifiedTime = observationList.getDatabaseTime();

		mObservationCache.setCurrent(siteName, year, month, lastModifiedTime);

		//
		// Add the observations into the cache.
		//
		mObservationCache.addObservationList(observationList);
	}


	/**
	 * Create the plot parameters data to be used by the plotting panel.
	 * @return
	 */
	private PlotParameters createPlotParameters()
	{
		PlotParameters plotParameters= new PlotParameters();

		plotParameters.mReport = mReport;

		plotParameters.mFullObservationList = mFullObservationList;
		plotParameters.mViewedObservationList = mViewedObservationList;
		plotParameters.mObservationList1 = mObservationList1;
		plotParameters.mObservationList2 = mObservationList2;

		plotParameters.mSelectedObservation = mSelectedObservation;
		plotParameters.mSelectedIsPlot1 = mSelectedIsPlot1;

		plotParameters.mMarkerObservation = mMarkerObservation;

		plotParameters.mSelector1 = mSelector1;
		plotParameters.mSelector2 = mSelector2;

		plotParameters.mTimeMajorInterval = mTimeMajorInterval;
		plotParameters.mTimeMinorInterval = mTimeMinorInterval;

		plotParameters.mMinimumTime = mMinimumTime;
		plotParameters.mMaximumTime = mMaximumTime;

		plotParameters.mShiftHours = mShiftHours;
		plotParameters.mTimeInterval = mTimeInterval;
		plotParameters.mGroupHours = mGroupHours;
		plotParameters.mIsUniformScaling = mIsUniformScaling;
		plotParameters.mIsShowUTCTimes = mIsShowUTCTimes;

		return plotParameters;
	}


	/**
	 * Get the time to be used for the newest time of the plot.
	 * This is the time of the latest observation for the site or the current time,
	 * whichever is later.  (The current time should really be later than the last
	 * observation time, but maybe the system's clock is off.)
	 */
	private DateTime calculatePlotTime(String siteName)
	{
		DateTime currentTime = new DateTime(new Date().getTime());

		Observation lastObservation = mReport.getObservationList().getLastObservation();

		if (lastObservation == null)
			return currentTime;

		DateTime lastTime = lastObservation.getUTCTime();

		if (lastTime.compareTo(currentTime) < 0)
			lastTime = currentTime;

		return lastTime;
	}


	/**
	 * Calculate all the time-related values to be used for the plot.
	 * This is dependent on the time interval, the shift value, and on
	 * the observations.
	 */
	private void calculateTimeValues()
	{
		//
		// If the time interval is special then calculate the intervals based
		// on just the observations from the current report or from all
		// of the observations for the site.
		//
		if ((mTimeInterval == 0) || (mTimeInterval == Long.MAX_VALUE))
		{
			calculateCurrentTimeValues();

			return;
		}

		//
		// The intervals are based on the specified time interval back
		// from the plot time.  Determine the major and minor time
		// intervals to be used for rounding.
		//
		int minutes = (int) (mTimeInterval / (DateTime.MINUTE_INTERVAL));

		calculateTimeIntervals(minutes);

		int hourMultiple = mTimeMajorInterval / 60;

		//
		// Now perform the rounding.
		// This is a little tricky because the rounding is done on the local
		// times of the observations, not the UTC time.  So we need to
		// examine the last observation to get its local time offset.
		//
		long timeOffset = 0;

		Observation lastObservation = mFullObservationList.getLastObservation();

		if (lastObservation != null)
			timeOffset = lastObservation.getLocalTimeOffset();

		DateTime localTime = mPlotTime.getAdjustedTime(timeOffset);

		localTime = localTime.roundUpToHourMultiple(hourMultiple);

		//
		// Finally convert the rounded time back to UTC time.
		//
		mMaximumTime = localTime.getAdjustedTime(-timeOffset);

		//
		// Now apply the shift to the maximum time and calculate the
		// corresponding minimum time from the time interval.
		//
		mMaximumTime = mMaximumTime.getAdjustedTime(-mShiftHours * DateTime.HOUR_INTERVAL);
		mMinimumTime = mMaximumTime.getAdjustedTime(-mTimeInterval);
	}


	/**
	 * Calculate the time-related values based on just the current observations
	 * to be plotted.  This is used when showing just the current report or all
	 * of the observations for the site.
	 */
	private void calculateCurrentTimeValues()
	{
		//
		// Calculate the amount of time the plot covers.
		//
		Observation firstObservation = mViewedObservationList.getFirstObservation();
		Observation lastObservation = mViewedObservationList.getLastObservation();

		if (firstObservation == null)
			return;

		DateTime firstTime = firstObservation.getUTCTime();
		DateTime lastTime = lastObservation.getUTCTime();

		int minutes = firstTime.getMinutesTo(lastTime);

		calculateTimeIntervals(minutes);

		int hourMultiple = mTimeMajorInterval / 60;

		//
		// Now perform the rounding.
		// The rounding is done based on the local time.
		//
		firstTime = firstObservation.getLocalTime();
		lastTime = lastObservation.getLocalTime();

		DateTime localMinimumTime = firstTime.roundDownToHourMultiple(hourMultiple);
		DateTime localMaximumTime = lastTime.roundUpToHourMultiple(hourMultiple);

		//
		// Finally convert the rounded times back to UTC time.
		//
		long firstTimeOffset = firstObservation.getLocalTimeOffset();
		long lastTimeOffset = lastObservation.getLocalTimeOffset();

		mMinimumTime = localMinimumTime.getAdjustedTime(-firstTimeOffset);
		mMaximumTime = localMaximumTime.getAdjustedTime(-lastTimeOffset);
	}


	/**
	 * Calculate the major and minor time intervals based on the specified
	 * number of minutes the plot will show.  The major time interval should be
	 * a multiple of an hour and a divisor of a day, or an exact multiple of days.
	 * The minor time interval should be a divisor of the major time interval and
	 * a multiple of 30 minutes.
	 * @param minutes
	 */
	private void calculateTimeIntervals(int minutes)
	{
		//
		// Calculate the major and minor intervals for the time grid.
		//
		TimeIntervals.IntervalInfo info = mTimeIntervals.getTimeInterval(minutes);

		if (info == null)
		{
			mTimeMajorInterval = 60 * 24;
			mTimeMinorInterval = 30 * 24;

			return;
		}

		mTimeMajorInterval = info.mMajorTime;
		mTimeMinorInterval = info.mMinorTime;
	}


	/**
	 * Find the marker observation to be used for the plot.
	 */
	private void findMarkerObservation()
	{
		if (mMarkerType.equals("Preserve"))
			return;

		if (mMarkerType.equals("Latest"))
		{
			mMarkerObservation = mFullObservationList.getLastObservation();

			return;
		}

		if (mMarkerType.equals("Selected"))
		{
			if (mSelectedIsPlot1 && (mSelectedObservation != null))
				mMarkerObservation = mSelectedObservation;

			return;
		}

		//
		// It's not one of the above, so assume it is "None".
		// Clear all markers.
		//
		mMarkerObservation = null;
	}


	/**
	 * Set whether uniform scaling is being used.
	 */
	public void setUniformScaling(boolean isUniformScaling)
	{
		mIsUniformScaling = isUniformScaling;

		updatePlot(false);
	}


	/**
	 * Set whether UTC times are being shown.
	 */
	public void setShowUTCTimes(boolean isShowUTCTimes)
	{
		mIsShowUTCTimes = isShowUTCTimes;

		updatePlot(true);
	}


	/**
	 * Set the marker type to be displayed.
	 */
	public void setMarkerType(String markerType)
	{
		mMarkerType = markerType;

		if (!markerType.equals("Preserve"))
			mMarkerObservation = null;

		updatePlot(false);
	}


	/**
	 * Set the plotting panel to display the values for the selected item.
	 * The selector indices are derived from the number keys of the keyboard.
	 * @param selectorIndex
	 */
	public void setSelector(int selectorIndex)
	{
		Selector selector = getSelector(selectorIndex);

		setSelector(selector);
	}


	/**
	 * Add the specified selector as the selector for a second plot.
	 * The selector indices are derived from the number keys of the keyboard.
	 * @param selectorIndex
	 */
	public void addSelector(int selectorIndex)
	{
		Selector selector = getSelector(selectorIndex);

		addSelector(selector);
	}


	/**
	 * Try to set the selector which specifies what is plotted.
	 * This clears the secondary selector unless we are in plot
	 * intervals mode, in which case the secondary selector
	 * is set to the same as the first selector.
	 * @param selector
	 */
	public void setSelector(Selector selector)
	{
		if ((selector != null) && selector.isPlottable())
			mSelector1 = selector;

		mSelector2 = null;

		if (mGroupHours > 0)
			mSelector2 = mSelector1;

		mDetailsDialog.setSelectors(mSelector1, mSelector2);

		//
		// Update the plot, but the times don't need changing.
		//
		updatePlot(false);
	}


	/**
	 * Try to set the secondary selector which adds a second plot.
	 * This doesn't change the primary selector.  This does nothing
	 * if we are doing an interval plot.
	 * @param selector
	 */
	public void addSelector(Selector selector)
	{
		if (mGroupHours > 0)
			return;

		if ((selector != null) && selector.isPlottable())
			mSelector2 = selector;

		if (mSelector2 == mSelector1)
			mSelector2 = null;

		mDetailsDialog.setSelectors(mSelector1, mSelector2);

		//
		// Update the plot, but the times don't need changing.
		//
		updatePlot(false);
	}


	/**
	 * Return the selector associated with the specified selector index
	 * (which is derived from a numeric key).  Returns null if the
	 * index is invalid.
	 */
	private Selector getSelector(int selectorIndex)
	{
		switch (selectorIndex)
		{
			case 1: return Selector.AIR_TEMP_SELECTOR;
			case 2: return Selector.APPARENT_AIR_TEMP_SELECTOR;
			case 3: return Selector.DEW_POINT_SELECTOR;
			case 4: return Selector.RELATIVE_HUMIDITY_SELECTOR;
			case 5: return Selector.DELTA_TEMP_SELECTOR;
			case 6: return Selector.WIND_SPEED_KMH_SELECTOR;
			case 7: return Selector.WIND_GUST_KMH_SELECTOR;
			case 8: return Selector.PRESSURE_QNH_SELECTOR;
			case 9: return Selector.PRESSURE_MSL_SELECTOR;
			case 0: return Selector.RAIN_SELECTOR;
		}

		return null;
	}


	/**
	 * Set the time interval in milliseconds.
	 * This is the maximum time between the first and last
	 * observation to be shown.  A value of zero indicates
	 * that only the current report's observations are shown.
	 * A value of MAX_VALUE indicates that all observations
	 * for the site are to be shown.
	 */
	public void setTimeInterval(long timeInterval)
	{
		mTimeInterval = timeInterval;
		mShiftHours = 0;

		updatePlot(true);
	}


	/**
	 * Shift the view leftwards by the specified number of hours.
	 * The value can be negative to shift the view rightwards.
	 * If there is no time interval then there is no shifting.
	 */
	public void shiftViewLeft(int hours)
	{
		if ((mTimeInterval == 0) || (mTimeInterval == Long.MAX_VALUE) ||
			(hours == 0))
		{
			return;
		}

		int newShiftHours = mShiftHours + hours;

		if (newShiftHours < 0)
			newShiftHours = 0;

		//
		// If the first observation is visible then no left shift can be done.
		//
		if ((hours > 0) &&
			(mMinimumTime.compareTo(mFullObservationList.getFirstObservation().getUTCTime()) <= 0))
		{
			return;
		}

		mShiftHours = newShiftHours;

		updatePlot(true);
	}


	public void showFirstObservation()
	{
		ObservationList observationList =
			(mSelectedIsPlot1 ? mObservationList1 : mObservationList2);

		selectObservation(mSelectedIsPlot1, observationList.getFirstObservation());
	}


	public void showLastObservation()
	{
		ObservationList observationList =
			(mSelectedIsPlot1 ? mObservationList1 : mObservationList2);

		selectObservation(mSelectedIsPlot1, observationList.getLastObservation());
	}


	public void showPreviousObservation()
	{
		ObservationList observationList =
			(mSelectedIsPlot1 ? mObservationList1 : mObservationList2);

		Observation observation = observationList.getPreviousObservation(mSelectedObservation);

		if (observation != null)
			selectObservation(mSelectedIsPlot1, observation);
	}


	public void showNextObservation()
	{
		ObservationList observationList =
			(mSelectedIsPlot1 ? mObservationList1 : mObservationList2);

		Observation observation = observationList.getNextObservation(mSelectedObservation);

		if (observation != null)
			selectObservation(mSelectedIsPlot1, observation);
	}


	/**
	 * Select the selected observation to be shown details of,
	 * or else to be removed if the observation is null.
	 * @param observation
	 */
	public void selectObservation(boolean isPlot1, Observation observation)
	{
		mSelectedObservation = observation;
		mSelectedIsPlot1 = isPlot1;

		mDetailsDialog.setObservation(observation);

		updatePlot(false);
	}


	/**
	 * Set the number of hours for an grouping plot as specified.
	 * The parameter can only have one of a few values.
	 * A value of zero disables the group plotting mode.
	 */
	public void setGroupHours(int groupHours)
	{
		mGroupHours = groupHours;

		updatePlot(true);
	}


	/**
	 * Show the actions dialog.
	 */
	public void showActionsDialog()
	{
		mActionsDialog.setVisible(true);
	}


	/**
	 * Exit from the application after cleaning up.
	 */
	public void exit()
	{
		LockManager.closeLock();

		System.exit(0);
	}


	/**
	 * Called when we are to run as a daemon application.
	 * We occasionally collect reports for all of the configured
	 * sites and save them all into the database, or else just do
	 * a single update and then exit.  This has no GUI to update.
	 */
	public void runDaemon(boolean isUpdate)
	{
		mPlotTime = new DateTime(new Date().getTime());

		//
		// Get the update interval and make sure it isn't too small.
		//
		long updateInterval = mConfig.getDaemonUpdateMinutes() * DateTime.MINUTE_INTERVAL;

		updateInterval = Math.max(updateInterval, 10 * DateTime.MINUTE_INTERVAL);

		//
		// Disable the time interval so that only the newest observations
		// are used.
		//
		mTimeInterval = 0;

		//
		// Get the set of sites to update.
		//
		Set<String> siteKeys = mConfig.getSiteMapKeys();

		while (true)
		{
			try
			{
				//
				// Update all of the site observations.
				//
				System.out.println("Collecting weather reports and updating the database");

				for (String siteKey : siteKeys)
				{
					String URL = mConfig.getSiteURL(siteKey);

					Report report = collectReport(URL);

					if (report != null)
						updateDatabase(report);
				}

				System.out.println("Update completed");

				//
				// If we were called to just do an update then exit now.
				//
				if (isUpdate)
				{
					System.out.println("Exiting");

					exit();
				}

				//
				// Wait a bit.
				//
				System.out.println("Sleeping");

				Thread.sleep(updateInterval);
			}
			catch (Exception e)
			{
				Utils.logError("Exception in daemon loop", e);

				exit();
			}
		}
	}


	/**
	 * Start the automatic updates timer if configured.
	 * This runs in a background thread when run as a normal GUI.
	 */
	public void startAutomaticUpdates()
	{
		int updateMinutes = mConfig.getUpdateMinutes();

		if (updateMinutes <= 0)
			return;

		if (updateMinutes > 60*24*30)
			updateMinutes = 60*24*30;

		mSleepTime = ((long) updateMinutes) * DateTime.MINUTE_INTERVAL;

		Runnable runnable = new Runnable()
		{
			public void run()
			{
				runAutomaticUpdateLoop();
			}
		};

		Thread thread = new Thread(runnable, "Updater");
		thread.setDaemon(true);

		thread.start();
	}


	/**
	 * Called from the separate automatic update thread to perform
	 * the automatic update loop.  This cannot directly update the
	 * GUI components since it is not the Swing thread so we have
	 * to call SwingUtils.
	 */
	public void runAutomaticUpdateLoop()
	{
		while (true)
		{
			try
			{
				Thread.sleep(mSleepTime);

				SwingUtilities.invokeAndWait(new Runnable()
				{
					public void run()
					{
						updateReport();
					}
				});
			}
			catch (Exception e)
			{
				Utils.logError("Exception in automatic update loop", e);

				return;
			}
		}
	}
}
