/**
 * Copyright (c) 2014 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather.gui;

import java.util.ArrayList;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.JTextField;

import weather.Configuration;
import weather.Controller;
import weather.Observation;

import weather.selector.Selector;


/**
 * A dialog which displays the details of a selected observation.
 */
public class DetailsDialog extends JDialog implements ActionListener
{
	private static final long serialVersionUID = 1L;

	private static final String ACTION_FIRST = "First";
	private static final String ACTION_LAST = "Last";
	private static final String ACTION_PREVIOUS = "Previous";
	private static final String ACTION_NEXT = "Next";

	private Controller mController;
	private Observation mObservation;

	private JPanel mButtonPanel;
	private JPanel mDataPanel;
	private JButton mFirstButton;
	private JButton mPrevButton;
	private JButton mNextButton;
	private JButton mLastButton;

	//
	// The primary and secondary fields currently selected (if any).
	//
	private FieldInfo mSelectedFieldInfo1;
	private FieldInfo mSelectedFieldInfo2;

	//
	// The colors to use for the selected fields.
	//
	private Color mSelectedColor1;
	private Color mSelectedColor2;

	/**
	 * The table of fields to be displayed in the dialog.
	 */
	private ArrayList<FieldInfo> mFieldInfoList = new ArrayList<FieldInfo>();


	/**
	 * Constructor for the dialog which specifies the owning frame
	 * and the application Controller.  The dialog is not yet shown.
	 * @param frame
	 * @param controller
	 */
	public DetailsDialog(Frame frame, Controller controller)
	{
		super(frame, "Observation Details");

		mController = controller;

		mController.setDetailsDialog(this);

		Configuration config = Configuration.getInstance();

		mSelectedColor1 = GUIUtils.getColorFromName(
			config.getPointColor(), Color.yellow);

		mSelectedColor2 = GUIUtils.getColorFromName(
			config.getPointColor2(), Color.pink);

		createDataPanel();
		createButtonPanel();

		Container content = getContentPane();

		content.setLayout(new BoxLayout(content, BoxLayout.Y_AXIS));

		content.add(mDataPanel);
		content.add(Box.createRigidArea(new Dimension(0, 5)));
		content.add(new JSeparator(JSeparator.HORIZONTAL));
		content.add(Box.createRigidArea(new Dimension(0, 5)));
		content.add(mButtonPanel);

		setLocation(500, 300);
		pack();

		//
		// Catch the window closing event.
		// We remove any selected observation in the plotting panel.
		//
		addWindowListener(new WindowAdapter()
			{
				@Override
				public void windowClosing(WindowEvent e)
				{
					mController.selectObservation(true, null);
				}
			});
	}


	/**
	 * Set or clear the observation to be shown in the details.
	 * The dialog is made visible or invisible accordingly.
	 */
	public void setObservation(Observation observation)
	{
		if (mObservation == observation)
			return;

		mObservation = observation;

		if (mObservation == null)
		{
			setVisible(false);

			return;
		}

		//
		// Set all the text field values for the observation.
		//
		for (FieldInfo fieldInfo : mFieldInfoList)
		{
			String text = fieldInfo.mSelector.getFormattedValue(mObservation);
			fieldInfo.mTextField.setText(text);
		}

		if (!isVisible())
			setVisible(true);
	}


	public Observation getObservation()
	{
		return mObservation;
	}


	/**
	 * Set the selectors whose rows are to show as being selected.
	 * The selectors can be null to remove any selection.
	 */
	public void setSelectors(Selector selector1, Selector selector2)
	{
		//
		// Clear any current selections.
		//
		setFieldColor(mSelectedFieldInfo1, null);
		setFieldColor(mSelectedFieldInfo2, null);

		//
		// Find the two fields being selected.
		//
		mSelectedFieldInfo1 = findFieldInfo(selector1);
		mSelectedFieldInfo2 = findFieldInfo(selector2);

		//
		// Set their colors.
		//
		setFieldColor(mSelectedFieldInfo1, mSelectedColor1);
		setFieldColor(mSelectedFieldInfo2, mSelectedColor2);
	}


	/**
	 * Find the field info object for the specified selector.
	 * Returns null if the field info was not found.
	 */
	private FieldInfo findFieldInfo(Selector selector)
	{
		if (selector == null)
			return null;

		for (FieldInfo fieldInfo : mFieldInfoList)
		{
			if (selector == fieldInfo.mSelector)
				return fieldInfo;
		}

		return null;
	}


	/**
	 * Set the label and test field background colors as specified.
	 * A null color sets to our default background color.
	 * The fieldInfo can be null in which case nothing is done.
	 */
	private void setFieldColor(FieldInfo fieldInfo, Color color)
	{
		if (fieldInfo == null)
			return;

		if (color == null)
			color = getBackground();

		fieldInfo.mLabel.setBackground(color);
		fieldInfo.mTextField.setBackground(color);
	}


	/**
	 * Create the button panel.
	 */
	private void createButtonPanel()
	{
		mButtonPanel = new JPanel(new FlowLayout());

		mFirstButton = createButton(ACTION_FIRST);
		mLastButton = createButton(ACTION_LAST);
		mPrevButton = createButton(ACTION_PREVIOUS);
		mNextButton = createButton(ACTION_NEXT);

		mButtonPanel.add(mFirstButton);
		mButtonPanel.add(mLastButton);
		mButtonPanel.add(mPrevButton);
		mButtonPanel.add(mNextButton);
	}


	private JButton createButton(String text)
	{
		JButton button = new JButton(text);

		button.setActionCommand(text);
		button.addActionListener(this);

		return button;
	}


	/**
	 * Create the data panel.  Each row contains a label and a text field.
	 */
	private void createDataPanel()
	{
		mDataPanel = new JPanel(new GridBagLayout());

		int row = 0;

		createDataRow(row++, Selector.SITE_NAME_SELECTOR);
		createDataRow(row++, Selector.UTC_TIME_SELECTOR);
		createDataRow(row++, Selector.LOCAL_TIME_SELECTOR);
		createDataRow(row++, Selector.SHORT_TIME_SELECTOR);
		createDataRow(row++, Selector.AIR_TEMP_SELECTOR);
		createDataRow(row++, Selector.APPARENT_AIR_TEMP_SELECTOR);
		createDataRow(row++, Selector.DEW_POINT_SELECTOR);
		createDataRow(row++, Selector.RELATIVE_HUMIDITY_SELECTOR);
		createDataRow(row++, Selector.DELTA_TEMP_SELECTOR);
		createDataRow(row++, Selector.PRESSURE_QNH_SELECTOR);
		createDataRow(row++, Selector.PRESSURE_MSL_SELECTOR);
		createDataRow(row++, Selector.WIND_SPEED_KMH_SELECTOR);
		createDataRow(row++, Selector.WIND_GUST_KMH_SELECTOR);
		createDataRow(row++, Selector.WIND_DIRECTION_SELECTOR);
		createDataRow(row++, Selector.RAIN_SELECTOR);

		//
		// Add mouse listeners to the created components.
		//
		addMouseListeners();
	}


	/**
	 * Create and add a label and text field to the dialog for the specified
	 * row and selector, and add a FieldInfo element for them.
	 * @param row
	 * @param selector
	 */
	private void createDataRow(int row, Selector selector)
	{
		GridBagConstraints c = new GridBagConstraints();

		c.gridx = 0;
		c.gridy = row;
		c.gridheight = 1;
		c.gridwidth = 1;
		c.weightx = 0;
		c.insets = new Insets(4, 4, 4, 4);
		c.fill = GridBagConstraints.NONE;

		JLabel label = new JLabel(selector.getName());
		label.setOpaque(true);

		mDataPanel.add(label, c);

		c.gridx = 1;
		c.weightx = 1;
		c.fill = GridBagConstraints.HORIZONTAL;

		JTextField textField = new JTextField(15);

		textField.setEditable(false);

		mDataPanel.add(textField, c);

		//
		// Add information about the details row to the list.
		//
		mFieldInfoList.add(new FieldInfo(label, textField, selector));
	}


	/**
	 * Add a mouse click listener to all the detail labels and text fields.
	 */
	private void addMouseListeners()
	{
		MouseListener listener = new MouseAdapter()
		{
			@Override
			public void mouseClicked(MouseEvent e)
			{
				detailsRowClicked(e);
			}
		};

		for (FieldInfo fieldInfo : mFieldInfoList)
		{
			fieldInfo.mLabel.addMouseListener(listener);
			fieldInfo.mTextField.addMouseListener(listener);
		}
	}


	/**
	 * Method called on a mouse click on a details row.
	 * This sets the selector for those details.
	 */
	private void detailsRowClicked(MouseEvent e)
	{
		int modifiers = e.getModifiersEx();
		Component component = e.getComponent();

		for (FieldInfo fieldInfo : mFieldInfoList)
		{
			if ((component == fieldInfo.mLabel) ||
				(component == fieldInfo.mTextField))
			{
				if ((modifiers & MouseEvent.CTRL_DOWN_MASK) == 0)
					mController.setSelector(fieldInfo.mSelector);
				else
					mController.addSelector(fieldInfo.mSelector);

				return;
			}
		}
	}


	/**
	 * Action handler for the buttons in the dialog.
	 */
	@Override
	public void actionPerformed(ActionEvent e)
	{
		String command = e.getActionCommand();

		if (command.equals(ACTION_FIRST))
			mController.showFirstObservation();

		if (command.equals(ACTION_LAST))
			mController.showLastObservation();

		if (command.equals(ACTION_PREVIOUS))
			mController.showPreviousObservation();

		if (command.equals(ACTION_NEXT))
			mController.showNextObservation();
	}


	/**
	 * A convenience class to hold information about one type of item.
	 */
	private static class FieldInfo
	{
		public final JLabel mLabel;
		public final JTextField mTextField;
		public final Selector mSelector;


		public FieldInfo(JLabel label, JTextField textField, Selector selector)
		{
			mLabel = label;
			mTextField = textField;
			mSelector = selector;
		}
	}
}

